local json = require( "json" )
require( "sqlite3" )

local function tableToString( t, indent )
	local str = ""
	if(indent == nil) then
		indent = 0
	end

	if(type(t) == "string") then
		str = str .. (" "):rep(indent) .. t .. "\n"
	elseif(type(t) == "number") then
		str = str .. (" "):rep(indent) .. t .. "\n"
	elseif(type(t) == "boolean") then
		if(t == true) then
			str = str .. "true"
		else
			str = str .. "false"
		end
	elseif(type(t) == "table") then
		local i, v
		for i, v in pairs(t) do
			if(type(v) == "table") then 
				str = str .. (" "):rep(indent) .. i .. "\n"
				str = str .. tableToString(v, indent + 2)
			else
				str = str .. (" "):rep(indent) .. i .. ": " .. tableToString(v, 0)
			end
		end
	end

	return str
end

local createHeader = function()
	local header = {}
	header.version = version
	header.created = os.time()
	header.modified = os.time()
	header.saved = os.time()
	return header
end

local openDatabase = function( databaseName )
	if databaseName then
		local path = system.pathForFile( databaseName, system.DocumentsDirectory )
		return sqlite3.open( path )
	end
end

local saveDatabase = function( database, items, header )
	if database then
		local query = [[CREATE TABLE IF NOT EXISTS icebox (id INTEGER PRIMARY KEY, value, header);]]
		database:exec( query )
		local encodedItems = json.encode( items or {} )
		if header then
			header.saved = os.time()
		end
		local encodedHeader = json.encode( header or createHeader() )
		local query = [[INSERT INTO icebox VALUES (NULL, ']] .. encodedItems .. [[',']] .. encodedHeader .. [['); ]]
		database:exec( query )
		database:close()
	end
end

local loadDatabase = function( database )
	local result = database:exec("SELECT * FROM icebox")
	local items = {}
	local header = {}

	if database then
		if result == 0 then
			for row in database:nrows("SELECT * FROM icebox") do
				local encodedItems = row.value
				items = json.decode( encodedItems )
				local encodedHeader = row.header
				header = json.decode( encodedHeader )
				if encodedHeader == "[]" then
					header = createHeader()
				end
			end
		end
	end
	return items, header
end

local deleteDatabase = function( databaseName )
	if databaseName then
		local path = system.pathForFile( databaseName, system.DocumentsDirectory )
		os.remove( path )
	end
end

local createRestorePointName = function( restorePointName, databaseName )
	if restorePointName and databaseName then
		return "ICE-RESTORE|" .. restorePointName .. "|" .. databaseName
	end
end

Ice = {}
Ice_mt = { __index = Ice }

function Ice:new()
	local self = {}
	self._boxes = {}
	setmetatable( self, Ice_mt )
	return self
end

function Ice:newBox( name )
	self:destroyBox( name )
	self._boxes[ #self._boxes + 1 ] = IceBox:new( name )
	return self._boxes[ #self._boxes ]
end

function Ice:getBox( name )
	for i = 1, #self._boxes, 1 do
		if self._boxes[i]:getName() == name then
			return self[ name ]
		end
	end
end

function Ice:destroyBox( name )
	local box = self:getBox( name )
	if box then
		box:destroy()
	end
end

function Ice:saveBox( name )
	local box = self:getBox( name )
	if box then
		box:save()
	end
end

function Ice:loadBox( name )
	local box = self:getBox( name )
	if not box then
		self._boxes[ #self._boxes + 1 ] = self:newBox( name )
	end
	self._boxes[ #self._boxes ]:load()
	self._boxes[ #self._boxes ]:save()
	return self._boxes[ #self._boxes ]
end

function Ice:storeBoxValue( boxName, valueName, value )
	local box = self:getBox( boxName )
	if box then
		box:store( valueName, value )
	end
end

function Ice:retrieveBoxValue( boxName, valueName )
	local box = self:getBox( boxName )
	if box then
		return box:retrieve( valueName )
	end
end

function Ice:save()
	for i = 1, #self._boxes, 1 do
		self._boxes[i]:save()
	end
end

function Ice:clear()
	for i = 1, #self._boxes, 1 do
		self._boxes[i]:clear()
	end
end

function Ice:delete()
	for i = 1, #self._boxes, 1 do
		self._boxes[i]:delete()
	end
end

function Ice:destroy()
	for i = 1, #self._boxes, 1 do
		self._boxes[i]:destroy()
	end
end

IceBox = {}
IceBox_mt = { __index = IceBox }

function IceBox:new( name )
	local self = {}
	self._items = {}
	self._version = 1.0
	self._name = name or "default"
	self._databaseName = self._name .. ".ice"
	setmetatable( self, IceBox_mt )
	return self
end

function IceBox:store( name, value )
	if name then
		self._items[ name ] = value
		if not self._header then
			self._header = createHeader()
		end
		self._header.modified = os.time()
	end
	if self:isAutomaticSavingEnabled() then
		self:save()
	end
	return value
end

function IceBox:storeIfNew( name, value )
	local added = false
	if name then
		if self:hasValue( name ) then
		else
			self._items[ name ] = value
			added = true
			self._header.modified = os.time()
		end
	end
	if self:isAutomaticSavingEnabled() then
		self:save()
	end
	return added, value
end

function IceBox:storeIfHigher( name, value )
	local added = false
	if name then
		local currentValue = self:retrieve( name ) or 0
		if currentValue then
			if value > currentValue then
				self._items[ name ] = value
				added = true
				self._header.modified = os.time()
			end
		else
			self._items[ name ] = value
			added = true
			self._header.modified = os.time()
		end
	end
	if self:isAutomaticSavingEnabled() then
		self:save()
	end
	return added, value
end

function IceBox:storeIfLower( name, value )
	local added = false
	if name then
		local currentValue = self:retrieve( name ) or 0
		if currentValue then
			if currentValue ~= 0 then
				if value < currentValue then
					self._items[ name ] = value
					added = true
					self._header.modified = os.time()
				end
			else
				self._items[ name ] = value
				added = true
				self._header.modified = os.time()
			end
		end
	end
	if self:isAutomaticSavingEnabled() then
		self:save()
	end
	return added, value
end

function IceBox:retrieve( name )
	if name then
		return self._items[ name ]
	end
end

function IceBox:increment( name, amount )
	local value = self:retrieve( name )
	if not value then
		value = self:store( name, 0 )
	end
	if value and type( value ) == "number" then
		value = value + ( amount or 1 )
	end
	self:store( name, value )
	self._header.modified = os.time()
	if self:isAutomaticSavingEnabled() then
		self:save()
	end
	return added, value
end

function IceBox:decrement( name, amount )
	local value = self:retrieve( name )
	if not value then
		value = self:store( name, 0 )
	end
	if value and type( value ) == "number" then
		value = value - ( amount or 1 )
	end
	self:store( name, value )
	self._header.modified = os.time()
	if self:isAutomaticSavingEnabled() then
		self:save()
	end
	return added, value
end

function IceBox:hasValue( name )
	if name then
		if self._items[ name ] ~= nil then
			return true
		else
			return false
		end
	end
end

function IceBox:remove( name )
	if name then
		if self:hasValue( name ) then
			self._items[ name ] = nil
			self._header.modified = os.time()
		end
	end
	if self:isAutomaticSavingEnabled() then
		self:save()
	end
	return added, value
end

function IceBox:getItems()
	return self._items
end

function IceBox:getHeader()
	return self._header
end

function IceBox:enableAutomaticSaving()
	self._automaticSavingEnabled = true
end

function IceBox:disbleAutomaticSaving()
	self._automaticSavingEnabled = false
end

function IceBox:isAutomaticSavingEnabled()
	return self._automaticSavingEnabled
end

function IceBox:save()
	self:delete()
	local database = openDatabase( self._databaseName )
	if database then
		saveDatabase( database, self._items, self._header )
	end
end

function IceBox:load()
	local database = openDatabase( self._databaseName )
	if database then
		self._items, self._header = loadDatabase( database )
		if not self._header or self._header == {} then
			self._header = createHeader()
		end
	end
end

function IceBox:getName()
	return self._name
end

function IceBox:print()
	print( tableToString( self._items ) )
end

function IceBox:setRestorePoint( name )
	if name then
		local restoreName = createRestorePointName( name, self._databaseName )
		local database = openDatabase( restoreName )
		if database then
			saveDatabase( database, self._items )
		end
	end
end

function IceBox:restoreToPoint( name )
	if name then
		local restoreName = createRestorePointName( name, self._databaseName )
		local database = openDatabase( restoreName )
		if database then
			self:clear()
			self._items, self._header = loadDatabase( database )
			if self:isAutomaticSavingEnabled() then
				self:save()
			end
		end
	end
end

function IceBox:deleteRestorePoint( name )
	if name then
		local restoreName = createRestorePointName( name, self._databaseName )
		deleteDatabase( restoreName )
	end
end

function IceBox:clear()
	self._items = nil
	self._items = {}
	if self:isAutomaticSavingEnabled() then
		self:save()
	end
end

function IceBox:delete()
	local path = system.pathForFile( self._databaseName, system.DocumentsDirectory )
	os.remove( path )
end

function IceBox:destroy()
	self:clear()
	self:delete()
end

_G.ice = Ice:new()
